/*********************************************************
 * Copyright (C) 2006 VMware, Inc. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation version 2.1 and no later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA.
 *
 *********************************************************/


/*
 * vmcheck.c --
 *
 *    Utility functions for discovering our virtualization status.
 */

#ifdef __cplusplus
extern "C" {
#endif

#include <stdlib.h>

#include "vmware.h"
#include "vm_version.h"
#include "vm_tools_version.h"
#include "backdoor.h"
#include "backdoor_def.h"
#include "debug.h"

#if !defined(_WIN32) && !defined(N_PLAT_NLM)
#   include "vmsignal.h"
#   include "setjmp.h"
#endif

#ifdef NETAPP
#if defined(LINUX)
#include <libnglinux/kenv_linux.h>
#include <libnglinux/linux_strings.h>
#else
#include <kenv.h>
#endif
#include "vmcheck.h"
#include "cpuid_info.h"
#include "string.h"
#endif /* NETAPP */

#if !defined(_WIN32) && !defined(N_PLAT_NLM)
static sigjmp_buf jmpBuf;
static Bool       jmpIsSet;


/*
 *----------------------------------------------------------------------
 *
 * VmCheckSegvHandler --
 *
 *    Signal handler for segv. Return to the program state saved
 *    by a previous call to sigsetjmp, or Panic if sigsetjmp hasn't
 *    been called yet. This function never returns;
 *
 * Return Value:
 *    None.
 *
 * Side effects:
 *    See the manpage for sigsetjmp for details.
 *
 *----------------------------------------------------------------------
 */

static void
VmCheckSegvHandler(int clientData) // UNUSED
{
   if (jmpIsSet) {
      siglongjmp(jmpBuf, 1);
   } else {
#if defined(LINUX)
#ifndef NETAPP
      Panic("Recieved SEGV, exiting");
#endif /* !NETAPP */
#endif /* LINUX */
   }
}
#endif


/*
 *----------------------------------------------------------------------
 *
 * VmCheck_GetVersion --
 *
 *    Retrieve the version of VMware that's running on the
 *    other side of the backdoor.
 *
 * Return value:
 *    TRUE on success
 *       *version contains the VMX version
 *       *type contains the VMX type
 *    FALSE on failure
 *
 * Side effects:
 *    None
 *
 *----------------------------------------------------------------------
 */

Bool
VmCheck_GetVersion(uint32 *version, // OUT
                    uint32 *type)    // OUT
{
   Backdoor_proto bp;

   ASSERT(version);
   ASSERT(type);

   /* Make sure EBX does not contain BDOOR_MAGIC */
   bp.in.size = ~BDOOR_MAGIC;
   /* Make sure ECX does not contain any known VMX type */
   bp.in.cx.halfs.high = 0xFFFF;

   bp.in.cx.halfs.low = BDOOR_CMD_GETVERSION;
   Backdoor(&bp);
   if (bp.out.ax.word == 0xFFFFFFFF) {
      /*
       * No backdoor device there. This code is not executing in a VMware
       * virtual machine. --hpreg
       */
      return FALSE;
   }

   if (bp.out.bx.word != BDOOR_MAGIC) {
      return FALSE;
   }

   *version = bp.out.ax.word;

   /*
    * Old VMXs (workstation and express) didn't set their type. In that case,
    * our special pattern will still be there. --hpreg
    */

   /* 
    * Need to expand this out since the toolchain's gcc doesn't like mixing
    * integral types and enums in the same trinary operator.
    */
   if (bp.in.cx.halfs.high == 0xFFFF)
      *type = VMX_TYPE_UNSET;
   else
      *type = bp.out.cx.word;

   return TRUE;
}

#ifdef NETAPP
/*
 * Identify the hypervisor based on the HYPERVISOR_CPUID input name
 */
Bool
identifyHypervisor(const char* name, char **hyperName, int size)
{
   if (0 == strncmp(XEN_STRING, name, sizeof(XEN_STRING))){
      strlcpy(*hyperName, "Xen", size);
      return TRUE;
   } else if (0 == strncmp(VMWARE_STRING, name, sizeof(VMWARE_STRING))) {
      strlcpy(*hyperName, "VMware", size);
      return TRUE;
   } else if (0 == strncmp(HYPERV_STRING, name, sizeof(HYPERV_STRING))) {
      strlcpy(*hyperName, "HyperV", size);
      return TRUE;
   } else if (0 == strncmp(KVM_STRING, name, sizeof(KVM_STRING))) {
      strlcpy(*hyperName, "KVM", size);
      return TRUE;
   } else {
      /* Unknown hypervisor string found  */
	  strlcpy(*hyperName, name, size);
   }
   return FALSE;
}


/*
 *----------------------------------------------------------------------
 *
 *  getHypervisor --
 *
 *      Check for for Hypervisor string.
 *
 *      Official way is to check CPUID function 0x4000 0000, which
 *         returns a hypervisor string.  See PR156185,
 *         http://xenbits.xensource.com/xen-unstable.hg?file/6a383beedf83/tools/misc/xen-detect.c
 *      (Canonical way is /proc/xen, but CPUID is better).
 *
 * Results:
 *      TRUE if we find a known hypervisor string.
 *      Returns hyperName of "Xen" if mathces XEN_STRING or "VMware" if it matches VMWARE_STRING.
 *      If hypervisor field is not NULL it returns FALSE and the value of field.
 *      Linux:
 *         Illegal instruction exception on real hardware?
 *         Some hypervisors implementations may not set string and return FALSE.
 *      Windows:
 *         FALSE on real hardware.
 *
 * Side effects:
 *	Linux: Will raise exception on native hardware?
 *	Windows: None.
 *
 *----------------------------------------------------------------------
 */

Bool
getHypervisor(char *hyperName, int size)
{
#define HYPERVISOR_CPUID 0x40000000
   CPUIDRegs regs;
   uint32 name[4];

   /*
    * HVM mode: simple cpuid instr
    * Xen and VMWare hypervisors trap CPUID and adds information leaf(s)
    * at CPUID leaf HYPERVISOR_CPUID and higher.
    */
   __GET_CPUID(HYPERVISOR_CPUID, &regs);
   name[0] = regs.ebx;
   name[1] = regs.ecx;
   name[2] = regs.edx;
   name[3] = 0;

   return identifyHypervisor((const char*)name, &hyperName, size);
}
#endif /* NETAPP */

/*
 *----------------------------------------------------------------------
 *
 * VmCheck_IsVirtualWorld --
 *
 *    Verify that we're running in a VM & we're version compatible with our
 *    environment.
 *
 * Return value:
 *    TRUE if we're in a virtual machine, FALSE otherwise.
 *
 * Side effects:
 *    None.
 *
 *----------------------------------------------------------------------
 */

Bool
VmCheck_IsVirtualWorld(void)
{
   uint32 version;
   uint32 dummy;

#ifdef NETAPP
   char env[64];

   /* Firstly check if it is running in VM or not */
   if (kenv(KENV_GET, "bootarg.vm", env, sizeof(env)) < 0) {
	   /* Not running in VM */
	   return FALSE;
   } else {
	   if (strcasecmp(env, "true") != 0) {
		   /* Not running in VM */
		   return FALSE;
	   }
   }

   /* If it's XEN we bail out */
   if (kenv(KENV_GET, "bootarg.vm.xen.vmtools_disable", env, sizeof(env)) >= 0) {
	   if (strcasecmp(env, "true") == 0) {
		   /* Bail out if it's XEN */
		   exit(1);
	   }
   }

   /*
    * Secondly check if it is running in VMware or not through smbios.
    * For VMware VM, we check smbios.system.serial with value like
    * "VMware-42 17 d1 0d ff e6 5f a7-1a 08 52 9c 3d ed 68 15"
    * For HyperV, we check smbios.system.maker with value like
    * "Microsoft Corporation".
    * Note that for VMware, smbios.system.maker comes from the real
    * vendor of physical host like smbios.system.maker="IBM", and
    * smbios.system.serial is the only smbios info for VMware check.
    */
   if (kenv(KENV_GET, "smbios.system.serial", env, sizeof(env)) < 0) {
	   return FALSE;
   } else {
	   if (strncasecmp(env, "VMware-", 7) != 0) {
                  /*
                   * BURT 852826:
                   *  In ESX 5.5U1 when vmware config SMBOIS.reflectHosts is set to TRUE,
                   * "VMware" is no longer in serial. Use cpuid check in getHypervisor()
                   */
                   if (getHypervisor(env, sizeof(env)) &&
                       strncasecmp(env, "VMware", 6) == 0) {
		       return TRUE;
                   } else {
		       /* Not running in VMware */
		       return FALSE;
                   }
	   }
   }
#endif /* NETAPP */

#ifdef N_PLAT_NLM
   /*
    * We are running at CPL0. So we'll not receive SIGSEGV on access
    * and we must do it other way... --petr
    */
   if (!VmCheck_GetVersion(&version, &dummy)) {
      return FALSE;
   }
#elif defined _WIN32
   __try {
      VmCheck_GetVersion(&version, &dummy);
   } __except (GetExceptionCode() == STATUS_PRIVILEGED_INSTRUCTION) {
      return FALSE;
   }
#else // POSIX
   int signals[] = {
      SIGSEGV,
   };
   struct sigaction olds[ARRAYSIZE(signals)];

   if (Signal_SetGroupHandler(signals, olds, ARRAYSIZE(signals),
                              VmCheckSegvHandler) == 0) {
      exit(1);
   }
   if (sigsetjmp(jmpBuf, TRUE) == 0) {
      jmpIsSet = TRUE;
      VmCheck_GetVersion(&version, &dummy);
   } else {
      jmpIsSet = FALSE;
      return FALSE;
   }

   if (Signal_ResetGroupHandler(signals, olds, ARRAYSIZE(signals)) == 0) {
      exit(1);
   }
#endif
   if (version != VERSION_MAGIC) {
#if defined(LINUX)
#ifndef NETAPP
      Debug("The version of this program is incompatible with your %s.\n"
            "For information on updating your VMware Tools please see\n"
            "http://www.vmware.com/support/reference/common/info_tools.html.\n"
            "\n", PRODUCT_LINE_NAME);
#endif /* !NETAPP */
#endif /* LINUX */
      return FALSE;
   }

   return TRUE;
}

#ifdef NETAPP /* burt 769059 */
Bool
doHostTimeSync(void)
{
    char val[8];
    int ret;

    ret = kenv(KENV_GET, "bootarg.vm.timesync", val, sizeof(val));
    if (ret == -1 || strcasecmp(val, "true") != 0) {
	return FALSE;
    } else {
	return TRUE;
    }
}
#endif /* NETAPP */

#ifdef __cplusplus
}
#endif
