#!/usr/bin/bash

#import common functions
. /etc/netapp_web_jail_common.subr

################################################################################
#Internal Functions
################################################################################

#Helper function to unmount directories that are mounted during setup
function unmountDependencies {
    umount -f $D/dev
    umount -f $D/lib
    umount -f $D/libexec
    umount -f $D/usr/lib
    umount -f $D/var/run
    umount -f $D/sbin
    umount -f $D/var/www/dev
    umount -f $D/var/www/svcs
    umount -f $D/mroot/etc/www/dev
    umount -f $D/mroot/etc/www/svcs
    umount -f $D/usr/www/svcs
    umount -f $D/etc/php5
    umount -f $D/var/webjail/etc
    umount -f $D/tmp/mntsvc
    umount -f $D/NDU
    umount -f $D/var/run

    umountCertificates $D false

    #unmount image file system
    umount -f $D$imageMountPoint
}

################################################################################
#Setup
################################################################################

#need root privileges to run
checkForSudo

#the /tmp directory will be limited to 5MB when on a vsim
#but will be 1GB when installed on real hardware
num1MBBlocks=5
if [ "$(kenv -q bootarg.vm)" != "true" ]; then
    num1MBBlocks=1024

    #if var isn't mounted yet, then this will cause some error messages to be
    #printed (see burt 1025925)
    if [ -e /var/etc/php.ini ]; then
        #we need to increase some of the php.ini file defaults so we can
        #support large uploads (such as an ONTAP image) in this increased space
        #these limits will remain 2M, 8M and 60 seconds on a vsim
        sed -Ei '' 's|(upload_max_filesize).+|\1 = 800M|' /var/etc/php.ini
        sed -Ei '' 's|(post_max_size).+|\1 = 1G|' /var/etc/php.ini
        sed -Ei '' 's|(max_input_time).+|\1 = 600|' /var/etc/php.ini
    fi
fi

#Read command line parameters and set options
readInputParameters PHP-FPM phpJail $num1MBBlocks "$@"

#make sure a jail isn't already running
if jailAlreadyRunning $D; then
    echo "There is already a jail running at $D"
    echo "Try shutting it down first (-s) or forcing a start (-f)"
    exit 1
fi

if [ $runlevel -ne 2 ]; then
    #create the main jail directory
    mkdirIfNotExist $D

    step "Mounting image filesystem"

    quietWrapper mountImgFilesystem $D $imageMountPoint $tmpSize $mdUnit

    #create the necessary directory structure for the jail
    step "Creating necessary directories"
    declare -a directories=(
        $D/mroot/etc/www/svcs $D/usr/www/svcs $D/var/www/svcs $D/var/run
        $D/libexec $D/dev $D/usr/lib $D/lib $D/sbin $D/etc/php5 $D/var/etc
        $D/var/webjail/etc $D/var/www/dev $D/mroot/etc/www/dev
        $D/NDU/mnt/store $D/tmp/upload
    )

    for dir in "${directories[@]}"; do
        mkdirIfNotExist $dir
    done

    #make tmp and upload dir writable by www user
    chown www $D/tmp
    chown www $D/tmp/upload

    makeCoreDir $D

    ############################################################################
    #Devices
    ############################################################################
    #We need to mount /dev into the jail using a devfs mount. There are only 2
    #devices needed for httpd to run correctly, urandom->random and null. We
    #will hide all of the other devices and explicilty show those 2.

    step "Creating device FS in jail"
    mountDevicesInJail $D

    ############################################################################
    #Libraries
    ############################################################################
    #We will be mounting the directories needed to supply shared libraries for
    #the httpd executable. This provides the benefit of not having 2+ copies of
    #the library (inside and outside of the jails) but the jail will have
    #visiblity of libraries and files placed in the /lib and /usr/lib
    #directories that it is not using.

    step "Mounting shared libraries"
    mountIfNotMounted /lib $D/lib
    mountIfNotMounted /usr/lib $D/usr/lib
    mountIfNotMounted /libexec $D/libexec

    ############################################################################
    #Binaries
    ############################################################################
    #We want to tightly control which executables the jail is allowed to run. To
    #do this, we supply it only with what it needs by copying them in instaed of
    #providing access to many more if we were to mount the directory.

    step "Mounting essential binaries"
    mountIfNotMounted /webjail/bin $D/sbin

    step "Mounting web services"
    mountIfNotMounted /usr/www/svcs $D/usr/www/svcs
    mountIfNotMounted /mroot/etc/www/svcs $D/mroot/etc/www/svcs
    mountIfNotMounted /var/www/svcs $D/var/www/svcs

    step "Mounting supdiag directories"
    mountIfNotMounted /var/www/dev $D/var/www/dev
    mountIfNotMounted /mroot/etc/www/dev $D/mroot/etc/www/dev

    step "Preparing for Automated NDU verifications"
    mkdirIfNotExist /mroot/etc/NDU/mnt/store
    mountIfNotMounted /mroot/etc/NDU $D/NDU
    mkdirIfNotExist $D/$D/NDU/mnt/store/etc/NDU/store
    version=$(uname -a | awk '{print $4}')
    ln -sf /NDU/mnt/store/etc/NDU/store/$version $D/$D/NDU/mnt/store/etc/NDU/store/$version
    ln -sf /NDU $D/mroot/etc/NDU

    ############################################################################
    #Config Files
    ############################################################################

    step "Setting up config files"

    #create and setup the webjail config directory (outside the jail)
    setupConfigFiles

    #Mount the webjail config directory
    mountIfNotMounted /var/webjail/etc $D/var/webjail/etc

    #Create symlinks to config files
    ln -sf /var/webjail/etc/php-fpm.conf $D/var/etc/php-fpm.conf
    ln -sf /var/webjail/etc/php.ini $D/var/etc/php.ini

    #create a hardlink to the hosts file so that it is part of the webjail folder
    ln -f /var/etc/hosts /var/webjail/etc/hosts
    ln -sf /var/webjail/etc/hosts $D/etc/hosts

    #its OK to copy in the services file because it can only change when the
    #system image changes and to do that this script will be run in the process
    cp /etc/services $D/etc/services

    #Mount the main configuration folder
    mountIfNotMounted /etc/php5 $D/etc/php5

    #Create the master.passwd file to make www a real user inside the jail
    printf "www:*:80:80::0:0:Jail Apache User:/nonexistent:/usr/sbin/nologin" > $D/var/etc/master.passwd
    #...and group
    printf "www:*:80:" > $D/etc/group

    #Generate the pwd database and the secure pwd database
    pwd_mkdb -d $D/var/etc $D/var/etc/master.passwd

    #Mount the directory containing the mlog socket for the jail
    mkdirIfNotExist /var/webjail/run
    ln -f /var/run/mlog /var/webjail/run/mlog
    mountIfNotMounted /var/webjail/run $D/var/run nullfs rw

    ############################################################################
    #File Upload Directory
    ############################################################################
    #make the upload directory for PHP file uploads
    mkdirIfNotExist $D/tmp/upload

    #create a link to the upload directory
    ln -sf $D/tmp/upload /mroot/etc/upload

    #change the permissions on /tmp and /tmp/upload
    chown www $D/tmp
    chown www $D/tmp/upload

    ############################################################################
    #Mount services
    ############################################################################
    #mount the mntsvc directory for SPI web services
    mkdirIfNotExist /tmp/mntsvc
    mkdirIfNotExist $D/tmp/mntsvc
    mountIfNotMounted /tmp/mntsvc $D/tmp/mntsvc

fi #end runlevel not equal 2

################################################################################
#Jail Command
################################################################################

if [ $runlevel -ge 2 ]; then
    step "Starting up jail"

    quietWrapper startupJail "$jailParent$jailName" "$D"

    #verify that the jail is now running
    runningId=$(getJailId $D)
    if [ -z $runningId ]; then
        quietWrapper unmountDependencies
        reportStartupError
        exit 1
    fi

fi #end runlevel >= 2
